# ----------------------------------------------------------------------
# Author:        yury.matveev@desy.de
# ----------------------------------------------------------------------

"""
"""

try:
    import PyTango
except ImportError:
    pass
import subprocess
import logging

import pyqtgraph as pg
from packaging import version

from PyQt5 import QtCore, QtWidgets
from distutils.util import strtobool

from petra_camera.utils.errors import report_error
from petra_camera.widgets.base_widget import BaseWidget
from petra_camera.main_window import APP_NAME
from petra_camera.gui.SettingsWidget_ui import Ui_SettingsWidget

from petra_camera.utils.functions import refresh_combo_box

WIDGET_NAME = 'SettingsWidget'

logger = logging.getLogger(APP_NAME)


# ----------------------------------------------------------------------
class SettingsWidget(BaseWidget):
    """
    """
    PARAMS_EDITOR = "atkpanel"

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        """
        """
        super(SettingsWidget, self).__init__(parent)

        self._ui = Ui_SettingsWidget()
        self._ui.setupUi(self)
        self._ui.gb_screen_motor.setVisible(self._camera_device.has_motor())

        # picture histogram
        if version.parse(pg.__version__) >= version.parse('0.12.0'):
            self.hist = pg.HistogramLUTWidget(self, orientation='horizontal')
        else:
            self.hist = pg.HistogramLUTWidget(self)

        self.hist.setBackground('w')
        self._ui.vl_levels.addWidget(self.hist, 1)

        self.hist.scene().sigMouseClicked.connect(self._hist_mouse_clicked)
        self.hist.item.sigLevelChangeFinished.connect(self.new_levels)
        self.hist.item.sigLookupTableChanged.connect(self.new_levels)
        self.hist.items()[-1].setMenuEnabled(False) # TODO better way

        # to prevent double code run
        self._my_mutex = QtCore.QMutex()

        self._load_camera_settings()

        # setup all signals
        self._ui.tbAllParams.clicked.connect(self._edit_all_params)

        for ui in ['exposure', 'gain', 'FPS', 'reduce']:
            getattr(self._ui, 'sb_{}'.format(ui)).editingFinished.connect(lambda x=ui: self._settings_changed(x))

        for ui in ['view_x', 'view_y', 'view_w', 'view_h']:
            getattr(self._ui, 'sb_{}'.format(ui)).editingFinished.connect(self._change_picture_size)

        self._ui.chk_auto_levels.stateChanged.connect(lambda state: self._switch_auto_levels(state == 2))
        self._ui.sb_max_level.valueChanged.connect(lambda value: self._level_settings_changed('level_max', value))
        self._ui.sb_min_level.valueChanged.connect(lambda value: self._level_settings_changed('level_min', value))
        self._ui.bg_level.buttonToggled.connect((lambda button: self._new_level_mode(button)))

        self._ui.cmb_path.currentTextChanged.connect(lambda text: self._camera_device.save_settings('Path', text))
        self._ui.cmb_source.currentTextChanged.connect(lambda text: self._camera_device.save_settings('Source', text))

        self._ui.but_in_out.clicked.connect(lambda: self._camera_device.move_motor())
        self._ui.chk_auto_screen.clicked.connect(lambda state: self._camera_device.set_auto_screen(state))

        self._ui.chk_dark_image.clicked.connect(lambda state: self._camera_device.toggle_dark_image(state==2))
        self._ui.but_acq_dark_image.clicked.connect(lambda: self._camera_device.set_dark_image())
        self._ui.but_save_dark_image.clicked.connect(self._save_dark_image)
        self._ui.but_load_dark_image.clicked.connect(self._load_dark_image)

        self._ui.chk_additional_settings.clicked.connect(lambda state: self._ui.gb_ext_settings.setVisible(state))

        self._ui.chk_background.clicked.connect(lambda state: self._camera_device.save_settings('background', state))
        self._ui.dsb_sigmas.valueChanged.connect(lambda value: self._camera_device.save_settings('background_sigmas', value))

    # ----------------------------------------------------------------------
    def refresh_view(self, force_read=False):
        """
        called periodically from camera window and syncs settings with Tango, etc...
        :return:
        """

        with QtCore.QMutexLocker(self._my_mutex):
            self._block_signals(True)

            self._ui.chk_auto_levels.setChecked(self._camera_device.levels['auto_levels'])
            self._ui.sb_min_level.setEnabled(not self._camera_device.levels['auto_levels'])
            self._ui.sb_max_level.setEnabled(not self._camera_device.levels['auto_levels'])

            if not self._ui.sb_min_level.hasFocus():
                self._ui.sb_min_level.setValue(self._camera_device.levels['levels'][0])
            if not self._ui.sb_max_level.hasFocus():
                self._ui.sb_max_level.setValue(self._camera_device.levels['levels'][1])

            self._ui.rb_lin_level.setChecked(self._camera_device.level_mode == 'lin')
            self._ui.rb_log_level.setChecked(self._camera_device.level_mode == 'log')
            self._ui.rb_sqrt_level.setChecked(self._camera_device.level_mode == 'sqrt')

            if self._ui.chk_additional_settings.isChecked() or force_read:
                if not self._ui.sb_exposure.hasFocus():
                    exposure_time = self._camera_device.get_settings('exposure', float)
                    self._ui.sb_exposure.setValue(exposure_time)

                if not self._ui.sb_gain.hasFocus():
                    gain_value = self._camera_device.get_settings('gain', int)
                    self._ui.sb_gain.setValue(gain_value)

                motor_position = self._camera_device.motor_position()
                if motor_position is not None:
                    self._ui.but_in_out.setText('Move Out' if motor_position else 'Move In')
                    self._ui.lb_screen_status.setText('Screen is In' if motor_position else 'Screen is Out')

                for ui in ['view_x', 'view_y']:
                    if not getattr(self._ui, 'sb_{}'.format(ui)).hasFocus():
                        getattr(self._ui, 'sb_{}'.format(ui)).setMaximum(1e6)
                        getattr(self._ui, 'sb_{}'.format(ui)).setValue(self._camera_device.get_settings(ui, int))

                for ui in ['view_w', 'view_h']:
                    if not getattr(self._ui, 'sb_{}'.format(ui)).hasFocus():
                        getattr(self._ui, 'sb_{}'.format(ui)).setMaximum(1e6)
                        value = self._camera_device.get_settings(ui, int)
                        if value == 0:
                            value = 1
                        getattr(self._ui, 'sb_{}'.format(ui)).setValue(value)

                self._update_picture_size_limits()

                if not self._ui.sb_FPS.hasFocus():
                    fps = self._camera_device.get_settings('FPS', int)
                    fps_max = self._camera_device.get_settings('FPSmax', int)
                    if fps == 0:
                        fps = 25
                    if fps_max == 0:
                        fps_max = 100

                    self._ui.sb_FPS.setMaximum(fps_max)
                    self._ui.sb_FPS.setValue(fps)

                if not self._ui.chk_background.hasFocus():
                    self._ui.chk_background.setChecked(self._camera_device.get_settings('background', bool))

                if not self._ui.dsb_sigmas.hasFocus():
                    self._ui.dsb_sigmas.setValue(self._camera_device.get_settings('background_sigmas', float))

                if not self._ui.sb_reduce.hasFocus():
                    self._ui.sb_reduce.setValue(self._camera_device.get_reduction())

                self._ui.chk_auto_screen.setChecked(self._camera_device.auto_screen)

                self._ui.but_save_dark_image.setEnabled(self._camera_device.has_dark_image())
                self._ui.chk_dark_image.setEnabled(self._camera_device.has_dark_image())
                self._ui.but_acq_dark_image.setEnabled(self._camera_device.got_first_frame)

                self._ui.chk_dark_image.setChecked(self._camera_device.subtract_dark_image)

            self._block_signals(False)

    # ----------------------------------------------------------------------
    def get_hist(self):
        return self.hist.item

    # ----------------------------------------------------------------------
    # --------------------- GUI  functionality  ----------------------------
    # ----------------------------------------------------------------------
    def _settings_changed(self, name):
        """
        general slot for arbitrary changed settings
        """
        if name == 'reduce':
            self._camera_device.set_reduction(self._ui.sb_reduce.value())

        self._camera_device.save_settings(name, getattr(self._ui, 'sb_{}'.format(name)).value())

    # ----------------------------------------------------------------------
    # ----------------- Histogram functionality ----------------------------
    # ----------------------------------------------------------------------
    def set_frame_to_hist(self, frame_view):
        """

        :param frame_view: picture form Frame viewer
        :return:
        """
        self.hist.item.setImageItem(frame_view)

    # ----------------------------------------------------------------------
    def block_hist_signals(self, flag):
        """

        :param flag: bool
        :return: None
        """
        if flag:
            self.hist.item.sigLevelChangeFinished.disconnect()
            self.hist.item.sigLookupTableChanged.disconnect()
        else:
            self.hist.item.sigLevelChangeFinished.connect(self.new_levels)
            self.hist.item.sigLookupTableChanged.connect(self.new_levels)

    # ----------------------------------------------------------------------
    def _switch_auto_levels(self, state):
        """
        slot for histogram sigLevelsChanged signal
        :return:
        """
        self._level_settings_changed('auto_levels', state)
        if state:
            self.hist.item.autoHistogramRange()

    # ----------------------------------------------------------------------
    def _hist_mouse_clicked(self, event):
        """
        double click turns on auto levels and does autorange
        :param event:
        :return:
        """
        if event.double():
            self._ui.chk_auto_levels.setChecked(True)

    # ----------------------------------------------------------------------
    def new_levels(self, lut_item):
        """
        slot for sigLookupTableChanged signal
        :param lut_item:
        :return:
        """
        lut = lut_item.saveState()
        lut['auto_levels'] = self._ui.chk_auto_levels.isChecked()

        self._camera_device.level_setting_change(lut)

    # ----------------------------------------------------------------------
    # --------------- Level settings ---------------------------------------
    # ----------------------------------------------------------------------
    def _level_settings_changed(self, param, value):
        """
        slots for levels adjustment functionality
        :param param: str, name of modified parameter
        :param value: new value
        :return: None
        """
        lev = list(self.hist.item.getLevels())
        if param == 'level_max':
            lev[1] = float(value)
        elif param == 'level_min':
            lev[0] = float(value)

        self.hist.item.setLevels(lev[0], lev[1])
        self.new_levels(self.hist.item)

    # ----------------------------------------------------------------------
    def _new_level_mode(self, button):
        """
        slot for level mode group
        :param button: selected button
        :return:
        """

        self._camera_device.set_new_level_mode(str(button.text()).lower())

    # ----------------------------------------------------------------------
    # ---------------------- Dark image functionality ----------------------
    # ----------------------------------------------------------------------
    def _save_dark_image(self):
        fileName, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Save dark image",
                                                            self._settings.option("save_folder", "default"),
                                                            "Numpy files (*.npy)")
        if fileName:
            self._camera_device.save_dark_image(fileName)

    # ----------------------------------------------------------------------
    def _load_dark_image(self):
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Load dark image",
                                                             self._settings.option("save_folder", "default"),
                                                             "Numpy files (*.npy)")
        if file_name:
            self._camera_device.load_dark_image(file_name)

    # ----------------------------------------------------------------------
    # ------------------- Picture clip -------------------------------------
    # ----------------------------------------------------------------------
    def _change_picture_size(self):

        self._camera_device.set_picture_clip(self._update_picture_size_limits())

    # ----------------------------------------------------------------------
    def _update_picture_size_limits(self):

        max_w, max_h = self._camera_device.get_max_picture_size()
        view_x = min(max(self._ui.sb_view_x.value(), 0), max_w)
        view_y = min(max(self._ui.sb_view_y.value(), 0), max_h)

        view_w = min(self._ui.sb_view_w.value(), max_w)
        if view_w == 0:
            view_w = 1
            self._ui.sb_view_w.blockSignals(True)
            self._ui.sb_view_w.setValue(view_w)
            self._ui.sb_view_w.blockSignals(False)

        view_h = min(self._ui.sb_view_h.value(), max_h)
        if view_h == 0:
            view_h = 1
            self._ui.sb_view_h.blockSignals(True)
            self._ui.sb_view_h.setValue(view_h)
            self._ui.sb_view_h.blockSignals(False)

        self._ui.sb_view_x.setMaximum(max_w - view_w)
        self._ui.sb_view_y.setMaximum(max_h - view_h)
        self._ui.sb_view_w.setMaximum(max_w - view_x)
        self._ui.sb_view_h.setMaximum(max_h - view_y)

        return view_x, view_y, view_w, view_h

    # ----------------------------------------------------------------------
    def _edit_all_params(self):
        """
        Open ATK panel for camera TODO fix it!
        """
        server = self._tangoServer
        server = "/" + "/".join(server.split("/")[1:])

        subprocess.Popen([self.PARAMS_EDITOR, server])

    # ----------------------------------------------------------------------
    def _block_signals(self, flag):
        """

        :param flag: bool
        :return: None
        """

        self._ui.sb_exposure.blockSignals(flag)
        self._ui.sb_gain.blockSignals(flag)

        self._ui.cmb_path.blockSignals(flag)

        self._ui.rb_lin_level.blockSignals(flag)
        self._ui.rb_log_level.blockSignals(flag)
        self._ui.rb_sqrt_level.blockSignals(flag)

    # ----------------------------------------------------------------------
    # --------------------Camera load and close functionality --------------
    # ----------------------------------------------------------------------
    def _load_camera_settings(self):

        result = True
        try:
            for layout in ['exposure', 'folder', 'FPS', 'source', 'background']:
                getattr(self._ui, 'frame_{}'.format(layout)).setVisible(layout in self._camera_device.visible_layouts())

            with QtCore.QMutexLocker(self._my_mutex):

                self._ui.cmb_path.clear()
                possible_folders = self._camera_device.get_settings('possible_folders', str)
                if possible_folders != '':
                    self._ui.cmb_path.setEnabled(True)
                    self._ui.cmb_path.addItems(possible_folders)
                    refresh_combo_box(self._ui.cmb_path, self._camera_device.get_settings('path', str))
                else:
                    self._ui.cmb_path.setEnabled(False)

                self._ui.cmb_source.clear()
                possible_sources = self._camera_device.get_settings('possible_sources', str)
                if possible_sources != '':
                    self._ui.cmb_source.setEnabled(True)
                    self._ui.cmb_source.addItems(possible_sources)
                    refresh_combo_box(self._ui.cmb_source, self._camera_device.get_settings('source', str))
                else:
                    self._ui.cmb_source.setEnabled(False)

                self.block_hist_signals(True)
                self.hist.item.restoreState(self._camera_device.levels)
                self.block_hist_signals(False)

            self.refresh_view(True)

        except Exception as err:
            report_error(err, self)
            result = False

        finally:
            self._block_signals(False)
            return result

    # ----------------------------------------------------------------------
    def load_ui_settings(self, camera_name):
        """
        Load basic GUI settings.

        :param camera_name:
        :return:
        """

        super(SettingsWidget, self).load_ui_settings(camera_name)

        settings = QtCore.QSettings(APP_NAME)

        state = False
        try:
            state = strtobool(settings.value(f"{WIDGET_NAME}_{camera_name}/AdditionalSettings"))
        except:
            pass

        self._ui.gb_ext_settings.setVisible(state)
        self._ui.chk_additional_settings.setChecked(state)

    # ----------------------------------------------------------------------
    def save_ui_settings(self, camera_name):
        """
        Save basic GUI settings.

        :param camera_name:
        :return:
        """
        super(SettingsWidget, self).save_ui_settings(camera_name)

        settings = QtCore.QSettings(APP_NAME)
        settings.setValue(f"{WIDGET_NAME}_{camera_name}/AdditionalSettings", self._ui.chk_additional_settings.isChecked())


# ----------------------------------------------------------------------
# Author:        yury.matveev@desy.de
# ----------------------------------------------------------------------

"""
This class provides data from camera, keeps, loads, saves camera settings etc
"""

import importlib
import logging
import threading
import time
import json
import math
import PyTango

import scipy.ndimage.measurements as scipymeasure
import numpy as np

try:
    from skimage.feature import peak_local_max
    peak_search = True
except:
    peak_search = False

from petra_camera.utils.errors import report_error
from petra_camera.utils.functions import FWHM
from petra_camera.main_window import APP_NAME

from PyQt5 import QtCore

logger = logging.getLogger(APP_NAME)


# ----------------------------------------------------------------------
class DataSource2D(QtCore.QObject):
    """
    """
    new_frame = QtCore.pyqtSignal()

    update_roi_statistics = QtCore.pyqtSignal()

    update_peak_search = QtCore.pyqtSignal()

    got_error = QtCore.pyqtSignal(str)

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        """
        """
        super(DataSource2D, self).__init__(parent)

        self._parent = parent
        self.settings = parent.settings

        self.device_id = ''
        self._base_id = ''

        self.auto_screen = False

        self._device_proxy = None
        self._worker = None  # thread, which reads from camera

        self._frame_mutex = QtCore.QMutex()  # sync access to frame
        self._last_frame = np.zeros((1, 1))  # keeps last read frame

        self.got_first_frame = False

        self._state = "idle"
        self.fps_limit = 1

        self.set_new_image = False

        # picture levels settings
        self.levels = []
        self.level_mode = 'lin'

        # ROIs parameters and data
        self.rois = []
        self.rois_data = []
        self._counter_roi = 0

        # ROIs parameters and data
        self.markers = []

        # peak search parameters and data
        self.peak_search = {}
        self.peak_coordinates = []

        self._dark_image = None
        self.subtract_dark_image = False

    # ----------------------------------------------------------------------
    def new_device_proxy(self, name, auto_screen):
        """

        :param name: str, camera name to be loaded from config
        :param auto_screen: bool, is generally moving screens is allowed
        :return: bool, success or not
        """

        for device in self.settings.get_nodes('camera_viewer', 'camera'):
            if device.getAttribute('name') == name:

                self._base_id = name

                try:
                    proxyClass = device.getAttribute("proxy")
                    logger.info("Loading device proxy {}...".format(proxyClass))

                    module = importlib.import_module("petra_camera.devices.{}".format(proxyClass.lower()))
                    self._device_proxy = getattr(module, proxyClass)(device)

                    self.device_id = self._base_id + self._device_proxy.file_name

                    # reset flags and variables
                    self.got_first_frame = False
                    self._last_frame = np.zeros((1, 1))

                    # load LUT and levels settings
                    lut = self.get_settings('lut', str)
                    if lut != '':
                        self.levels = json.loads(lut)
                    else:
                        self.levels = {'gradient': {'mode': 'rgb',
                                                    'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (255, 255, 255, 255))],
                                                    'ticksVisible': True},
                                       'levels': (0, 255.0),
                                       'mode': 'mono',
                                       'auto_levels': True}

                    self.level_mode = self.get_settings('level_mode', str)
                    if self.level_mode == '':
                        self.level_mode = 'lin'

                    self.auto_screen = self.get_settings('auto_screen', bool)

                    # load ROI params
                    self.rois = []
                    self.rois_data = []

                    # TODO find a better solution for Sardana counter....
                    self._counter_roi = self.get_settings('counter_roi', int)
                    self._counter_param = self.get_settings('counter_roi', int)

                    for ind in range(self.get_settings('num_rois', int)):
                        self.rois.append({'x': self.get_settings('roi_{}_x'.format(ind), int),
                                          'y': self.get_settings('roi_{}_y'.format(ind), int),
                                          'w': self.get_settings('roi_{}_w'.format(ind), int),
                                          'h': self.get_settings('roi_{}_h'.format(ind), int),
                                          'bg': self.get_settings('roi_{}_bg'.format(ind), int),
                                          'visible': self.get_settings('roi_{}_visible'.format(ind), bool),
                                          'mark': self.get_settings('roi_{}_mark'.format(ind), str),
                                          'color': self.get_settings('roi_{}_color'.format(ind), str)})

                        if ind == self._counter_roi:
                            for setting in ['x', 'y', 'w', 'h']:
                                self.save_settings('counter_{}'.format(setting), self.rois[ind][setting])

                        self.rois_data.append(dict.fromkeys(['max_x', 'max_y', 'max_v',
                                                             'min_x', 'min_y', 'min_v',
                                                             'com_x', 'com_y', 'com_v',
                                                             'fwhm_x', 'fwhm_y', 'sum']))

                    # load markers params
                    self.markers = []
                    for ind in range(self.get_settings('num_markers', int)):
                        self.markers.append({'x': self.get_settings('marker_{}_x'.format(ind), int),
                                             'y': self.get_settings('marker_{}_y'.format(ind), int),
                                             'visible': self.get_settings('marker_{}_visible'.format(ind), bool),
                                             'color': self.get_settings('marker_{}_color'.format(ind), str)})

                    # load peak search params
                    self.peak_search = {'search': False, #TODO: need solution for wrong settings!
                                        'search_mode': self.get_settings('peak_search_mode', bool),
                                        'rel_threshold': self.get_settings('peak_rel_threshold', int),
                                        'abs_threshold': self.get_settings('peak_abs_threshold', int)}

                    if self.peak_search['rel_threshold'] == 0:
                        self.peak_search['rel_threshold'] = 80

                    if self.peak_search['abs_threshold'] == 0:
                        self.peak_search['abs_threshold'] = 16000

                    # if Tango server for camera already acquiring - start data thread
                    if self._device_proxy.is_running():
                        self.start(auto_screen)

                    return True, ''

                except PyTango.DevFailed as ex:
                    logger.error(ex.args[0].desc)
                    return False, ex.args[0].desc

                except Exception as ex:
                    logger.error(ex)
                    return False, ex.__repr__()

        return False, 'Cannot find camera in config'

    # ----------------------------------------------------------------------
    def close_camera(self):
        """
        safe close for camera (e.g. to kill all threads)
        :return: None
        """

        if self._device_proxy is not None:
            self._device_proxy.close_camera()

    # ----------------------------------------------------------------------
    def start(self, auto_screen):
        """

        :param auto_screen: is generally moving screens is allowed
        :return:
        """
        self._reset_worker()
        self._worker.start()

        if self.auto_screen and auto_screen:
            self._device_proxy.move_motor(True)

        logger.debug('CameraDevice started')

    # ----------------------------------------------------------------------
    def stop(self, auto_screen):
        """

        :param auto_screen: is generally moving screens is allowed
        :return:
        """
        if self._state != 'idle':
            self._state = "abort"

        while self._state != 'idle':
            time.sleep(1e-3)

        if self.auto_screen and auto_screen:
            self._device_proxy.move_motor(False)

        logger.debug('CameraDevice stopped')

    # ----------------------------------------------------------------------
    # --------------------- Data acquiring thread---------------------------
    # ----------------------------------------------------------------------
    def _reset_worker(self):
        """
        """
        if self._worker:
            self._state = 'abort'
            self._worker.join()

        self._worker = threading.Thread(target=self.run)

    # ----------------------------------------------------------------------
    def run(self):
        """
        main thread cycle
        :return:
        """
        if self._start_acquisition():

            while self._state == "running":

                frame = self._device_proxy.maybe_read_frame()

                if frame is not None:
                    self.got_first_frame = True
                    self._last_frame = frame
                    self.new_frame.emit()

                    self.calculate_roi_statistics()
                    self.find_peaks()

                if self._device_proxy.error_flag:
                    self.got_error.emit(str(self._device_proxy.error_msg))
                    self._state = "abort"

                time.sleep(1 / self.fps_limit)  # to decrease processor load

            logger.info("Closing {}...".format(self.device_id))

            if self._device_proxy:
                self._device_proxy.stop_acquisition()

        self._state = "idle"

    # ----------------------------------------------------------------------
    def _start_acquisition(self):
        """
        tries to start camera
        :return: bool, success or not
        """

        if self._device_proxy:

            try:
                if self._device_proxy.start_acquisition():
                    self._state = "running"
                    return True
                else:
                    return False

            except Exception as err:
                report_error(err, self._parent)
                return False

    # ----------------------------------------------------------------------
    # ---------------------- Frame functionality ---------------------------
    # ----------------------------------------------------------------------
    def get_frame(self):
        """
        returns last frame after applying dark image and level mode
        :return: 2d np.array
        """
        if self.subtract_dark_image and self._dark_image is not None:
            try:
                invalid_idx = self._last_frame < self._dark_image
                frame = self._last_frame - self._dark_image
                frame[invalid_idx] = 0
            except:
                self.subtract_dark_image = False
                self._dark_image = None
                frame = self._last_frame
        else:
            frame = self._last_frame

        if self.level_mode == 'sqrt':
            frame = np.sqrt(np.abs(frame))
        elif self.level_mode == 'log':
            frame = np.log(np.maximum(1, frame))

        if np.max(frame) == 0:
            return np.ones_like(frame)
        else:
            return frame

    # ----------------------------------------------------------------------
    # ------------------- Levels functionality ----------------------------
    # ----------------------------------------------------------------------
    def level_setting_change(self, lut_state):
        """

        :param lut_state: dict, levels settings
        :return: None
        """
        lut_state['levels'] = (int(lut_state['levels'][0]), int(lut_state['levels'][1]))
        self.levels = lut_state
        self.save_settings('lut', json.dumps(lut_state))
        if self.got_first_frame:
            self.new_frame.emit()

    # ----------------------------------------------------------------------
    def set_new_level_mode(self, mode):
        """

        :param mode: srt, mode: "lin", "sqrt", "log"
        :return:
        """
        if self.level_mode != mode:
            self.level_mode = mode
            self.save_settings('level_mode', mode)
            if self.got_first_frame:
                self.new_frame.emit()

    # ----------------------------------------------------------------------
    # ------------------- Markers functionality ----------------------------
    # ----------------------------------------------------------------------
    def _save_marker_settings(self):
        """

        :return: drops markers to settings
        """
        num_markers = len(self.markers)
        self.save_settings('num_markers', num_markers)
        for ind, marker in enumerate(self.markers):
            for param in ['x', 'y', 'visible', 'color']:
                self.save_settings('marker_{}_{}'.format(ind, param), marker[param])

    # ----------------------------------------------------------------------
    def set_marker_value(self, marker_id, setting, value):
        """

        :param marker_id: int
        :param setting: str, setting name
        :param value: new value
        :return:
        """
        self.markers[marker_id][setting] = value
        self.save_settings('marker_{}_{}'.format(marker_id, setting), value)

    # ----------------------------------------------------------------------
    def append_marker(self):
        """

        :return: None
        """
        self.markers.append({'x': 0, 'y': 0, 'visible': True, 'color': self.settings.option('colors', 'marker')})
        self._save_marker_settings()

    # ----------------------------------------------------------------------
    def delete_marker(self, index):
        """

        :param index: int, maker to be deleted
        :return:
        """

        del self.markers[index]
        self._save_marker_settings()

    # ----------------------------------------------------------------------
    # ----------------------- ROI functionality ----------------------------
    # ----------------------------------------------------------------------
    def _save_roi_settings(self):
        """
        drops ROIs to settings
        :return:
        """
        num_rois = len(self.rois)
        self.save_settings('num_rois', num_rois)
        for ind, roi in enumerate(self.rois):
            for param in ['x', 'y', 'w', 'h', 'bg', 'visible', 'mark']:
                self.save_settings('roi_{}_{}'.format(ind, param), roi[param])

    # ----------------------------------------------------------------------
    def set_roi_value(self, roi_id, setting, value):
        """

        :param roi_id: int
        :param setting: str, setting name
        :param value:new value
        :return:
        """
        self.rois[roi_id][setting] = value
        self.save_settings('roi_{}_{}'.format(roi_id, setting), value)

        if roi_id == self._counter_roi:
            if setting in ['x', 'y', 'w', 'h']:
                self.save_settings('counter_{}'.format(setting), value)

    # ----------------------------------------------------------------------
    def num_roi(self):
        """
        return how many ROIs are there
        :return: int
        """
        return len(self.rois)

    # ----------------------------------------------------------------------
    def get_counter_roi(self):
        """

        :return: int, which ROIs selected as counter for Sardana
        """
        return self._counter_roi

    # ----------------------------------------------------------------------
    def set_counter_roi(self, value):
        """
        set new ROIs as counter for Sardana
        :param value: int
        :return:
        """
        self._counter_roi = value
        self.save_settings('counter_roi', value)
        for setting in ['x', 'y', 'w', 'h']:
            if len(self.rois):
                self.save_settings('counter_{}'.format(setting), self.rois[value][setting])

    # ----------------------------------------------------------------------
    def get_active_roi_value(self, value):
        """

        :param value: str, name of parameter
        :return:
        """
        return self.rois_data[self._counter_roi][value]

    # ----------------------------------------------------------------------
    def add_roi(self):
        """

        :return: None
        """
        self.rois.append({'x': 0, 'y': 0, 'w': 50, 'h': 50, 'bg': 0, 'visible': True, 'mark': '',
                          'color': self.settings.option('colors', 'roi')})
        self.rois_data.append(dict.fromkeys(['max_x', 'max_y', 'max_v',
                                             'min_x', 'min_y', 'min_v',
                                             'com_x', 'com_y', 'com_v',
                                             'fwhm_x', 'fwhm_y', 'sum']))

        self._save_roi_settings()

    # ----------------------------------------------------------------------
    def delete_roi(self, index):
        """

        :param index: roi to be deleted
        :return: None
        """

        del self.rois[index]
        del self.rois_data[index]
        self._save_roi_settings()

    # ----------------------------------------------------------------------
    def calculate_roi_statistics(self):
        """
        calculates ROIs statistics after new frame comes of ROIs parameter changed

        :return: None
        """
        if self._last_frame is None:
            return

        for info, data in zip(self.rois, self.rois_data):
            if info['visible']:

                image_size = self.get_picture_clip()
                _image_x_pos, _image_y_pos = image_size[0], image_size[1]
                x, y, w, h = int(info['x'] - _image_x_pos), int(info['y'] - _image_y_pos), int(info['w']), int(info['h'])

                array = self._last_frame[x:x + w, y:y + h]
                if array != []:
                    array[array < info['bg']] = 0  # All low values set to 0

                    roi_sum = np.sum(array)

                    try:
                        roiExtrema = scipymeasure.extrema(array)  # all in one!
                    except:
                        roiExtrema = (0, 0, (0, 0), (0, 0))

                    roi_max = (int(roiExtrema[3][0] + x + _image_x_pos), int(roiExtrema[3][1] + y + _image_y_pos))
                    roi_min = (int(roiExtrema[2][0] + x + _image_x_pos), int(roiExtrema[2][1] + y + _image_y_pos))

                    try:
                        roi_com = scipymeasure.center_of_mass(array)
                    except:
                        roi_com = (0, 0)

                    if math.isnan(roi_com[0]) or math.isnan(roi_com[1]):
                        roi_com = (0, 0)

                    roi_com = (int(roi_com[0] + x + _image_x_pos), int(roi_com[1] + y + _image_y_pos))

                    try:
                        intensity_at_com = self._last_frame[int(round(roi_com[0])), int(round(roi_com[1]))]
                    except:
                        intensity_at_com = [0, 0]

                    roi_FWHM = (FWHM(np.sum(array, axis=1)), FWHM(np.sum(array, axis=0)))

                    data['max_x'], data['max_y'] = roi_max
                    data['max_v'] = np.round(roiExtrema[1], 3)

                    data['min_x'], data['min_y'] = roi_min
                    data['min_v'] = np.round(roiExtrema[0], 3)

                    data['com_x'], data['com_y'] = roi_com
                    data['com_v'] = np.round(intensity_at_com, 3)

                    data['fwhm_x'], data['fwhm_y'] = roi_FWHM
                    data['sum'] = np.round(roi_sum, 3)

        self.update_roi_statistics.emit()

    # ----------------------------------------------------------------------
    # ------------- Peak search functionality ------------------------------
    # ----------------------------------------------------------------------
    def set_peak_search_value(self, setting, value):
        """

        :param setting: str, parameter name
        :param value: new value
        :return:
        """
        self.peak_search[setting] = value
        self.save_settings('peak_{}'.format(setting), value)
        self.find_peaks()

    # ----------------------------------------------------------------------
    def find_peaks(self):
        """
        finds peaks after new frame comes of parameter changed
        :return: None
        """
        if peak_search:
            if self.peak_search['search']:
                try:
                    if self.peak_search['search_mode']:
                        coordinates = peak_local_max(self._last_frame,
                                                     threshold_rel=self.peak_search['rel_threshold'] / 100)
                    else:
                        coordinates = peak_local_max(self._last_frame,
                                                     threshold_abs=self.peak_search['abs_threshold'])

                    if len(coordinates) > 100:
                        report_error(
                            'Too many ({}) peaks found. Show first 100. Adjust the threshold'.format(len(coordinates)),
                            self, True)
                        self.peak_coordinates = coordinates[:100]
                except:
                    self.peak_coordinates = ()
            else:
                self.peak_coordinates = ()
        else:
            self.peak_coordinates = ()

        self.update_peak_search.emit()

    # ----------------------------------------------------------------------
    # ---------------Communication with camera worker-----------------------
    # ----------------------------------------------------------------------
    def get_settings(self, setting, cast):
        """

        :param setting: str, settings name
        :param cast: expected type
        :return: cast(value), if there is no device proxy - None
        """

        if self._device_proxy:
            if setting == 'FPS':
                self.fps_limit = max(1, self._device_proxy.get_settings('FPS', int))
                return self.fps_limit
            else:
                return self._device_proxy.get_settings(setting, cast)
        else:
            return None

    # ----------------------------------------------------------------------
    def save_settings(self, setting, value):
        """

        :param setting: str, settings name
        :param value: value to save
        :return:
        """
        if self._device_proxy:
            if setting == 'FPS':
                self.fps_limit = value

            self._device_proxy.save_settings(setting, value)

    # ----------------------------------------------------------------------
    def is_running(self):
        """

        :return: bool, if there is no device proxy - None
        """
        if self._device_proxy is None:
            return None
        else:
            return self._device_proxy.is_running()

    # ----------------------------------------------------------------------
    def has_motor(self):
        """

        :return: bool
        """
        return self._device_proxy.has_motor()

    # ----------------------------------------------------------------------
    def move_motor(self, new_state=None):
        """

        :param new_state: True - move screen in, False - out
        :return:
        """

        self._device_proxy.move_motor(new_state)

    # ----------------------------------------------------------------------
    def motor_position(self):
        """

        :return: bool, True - move in, False - out; if there is no device proxy - None
        """
        if self._device_proxy is not None:
            return self._device_proxy.motor_position()

        return None

    # ----------------------------------------------------------------------
    def has_counter(self):
        """

        :return: bool, there is a corresponding Tango server or not
        """
        if self._device_proxy is not None:
            return self._device_proxy.has_counter()

        return False

    # ----------------------------------------------------------------------
    def get_counter(self):
        """

        :return: name of currently selected parameter in server
        """
        if self._device_proxy is not None:
            return self._device_proxy.get_counter()

        return ""

    # ----------------------------------------------------------------------
    def set_counter(self, value):
        """
        set new paramater in server
        :param value: str, parameter name
        :return:
        """
        self._device_proxy.set_counter(value)

    # ----------------------------------------------------------------------
    def visible_layouts(self):

        return self._device_proxy.visible_layouts

    # ----------------------------------------------------------------------
    def set_picture_clip(self, size):

        self.set_new_image = True
        self.save_settings('view_x', size[0])
        self.save_settings('view_y', size[1])
        self.save_settings('view_w', size[2])
        self.save_settings('view_h', size[3])
        self._device_proxy.set_picture_clip(size)

    # ----------------------------------------------------------------------
    def get_picture_clip(self):
        """

        :return: (x, y, w, h), where x, y - top left corner coordinates
        """
        return self._device_proxy.get_picture_clip()

    # ----------------------------------------------------------------------
    def get_max_picture_size(self):
        """

        :return: int, int, max frame width and heigth
        """
        return self._device_proxy.get_settings('max_width', int), self._device_proxy.get_settings('max_height', int)

    # ----------------------------------------------------------------------
    def get_reduction(self):
        """

        :return: int, reduction of camera resolution rate
        """
        return self._device_proxy.get_reduction()

    # ----------------------------------------------------------------------
    def set_reduction(self, value):

        self._device_proxy.set_reduction(value)
        self.set_new_image = True
        if self.got_first_frame:
            self.new_frame.emit()

    # ----------------------------------------------------------------------
    def set_auto_screen(self, state):
        """
         switch on or off screen motor
        :param state: bool
        :return:
        """
        self.auto_screen = state
        self.save_settings('auto_screen', state)

    # ----------------------------------------------------------------------
    # -------------------- Dark image functionality ------------------------
    # ----------------------------------------------------------------------
    def set_dark_image(self):
        """
        saves current image as dark image
        :return:
        """
        self._dark_image = self._last_frame

    # ----------------------------------------------------------------------
    def load_dark_image(self, file_name):
        """
        load dark image from file
        :param file_name:
        :return:
        """
        self._dark_image = np.load(file_name)

    # ----------------------------------------------------------------------
    def save_dark_image(self, file_name):
        """
        saves current image to the file
        :param file_name:
        :return:
        """
        np.save(file_name, self._dark_image)

    # ----------------------------------------------------------------------
    def has_dark_image(self):
        """

        :return: bool
        """
        return self._dark_image is not None

    # ----------------------------------------------------------------------
    def toggle_dark_image(self, state):
        """

        :param state: bool
        :return:
        """
        if self._dark_image is not None:
            self.subtract_dark_image = state

        if self.got_first_frame:
            self.new_frame.emit()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/MainWindow.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(845, 480)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 845, 21))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setSizeGripEnabled(False)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.quit_action = QtWidgets.QAction(MainWindow)
        self.quit_action.setObjectName("quit_action")
        self.actionAbout = QtWidgets.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.quit_action.setText(_translate("MainWindow", "Quit..."))
        self.quit_action.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.actionAbout.setText(_translate("MainWindow", "About..."))

import petra_camera.gui.icons_rc
